﻿using System;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.MultiObjectiveTestFunctions;

namespace HeuristicLab.Problems.MultiObjectiveTestFunction {
  [StorableClass]
  public class MultiObjectiveTestFunctionProblem : MultiObjectiveBasicProblem<RealVectorEncoding> {


    //TODO update of Maximisatzion when SolutionSize or TestFunction changes
    public override bool[] Maximization {
      get {
        return Parameters.ContainsKey("TestFunction") ? TestFunction.Maximization : new bool[2];
      }
    }

    #region Parameter Properties

    /// <summary>
    /// The dimensionality of the solution candidates
    /// </summary>
    private IFixedValueParameter<IntValue> ProblemSizeParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters["ProblemSize"]; }
    }

    /// <summary>
    /// The number of objectives that are to be optimized
    /// </summary>
    private IFixedValueParameter<IntValue> SolutionSizeParameter {
      get { return (IFixedValueParameter<IntValue>)Parameters["SolutionSize"]; }
    }

    /// <summary>
    /// The bounds for the entries of the solution candidate
    /// </summary>
    private IValueParameter<DoubleMatrix> BoundsParameter {
      get { return (IValueParameter<DoubleMatrix>)Parameters["Bounds"]; }
    }

    /// <summary>
    /// The testfunction 
    /// </summary>
    public IValueParameter<IMultiObjectiveTestFunction> TestFunctionParameter {
      get { return (IValueParameter<IMultiObjectiveTestFunction>)Parameters["TestFunction"]; }
    }
    #endregion

    #region Properties
    public int ProblemSize {
      get { return ProblemSizeParameter.Value.Value; }
      set { ProblemSizeParameter.Value.Value = value; }
    }
    public int SolutionSize {
      get { return SolutionSizeParameter.Value.Value; }
      set { SolutionSizeParameter.Value.Value = value; }
    }
    public DoubleMatrix Bounds {
      get { return BoundsParameter.Value; }
      set { BoundsParameter.Value = value; }
    }
    public IMultiObjectiveTestFunction TestFunction {
      get { return TestFunctionParameter.Value; }
      set { TestFunctionParameter.Value = value; }
    }
    #endregion

    public override void Analyze(Individual[] individuals, double[][] qualities, ResultCollection results, IRandom random) {
      base.Analyze(individuals, qualities, results, random);
      if (qualities[0].Length != 2) { throw new Exception(); }
      if (!results.ContainsKey("Hypervolume")) {
        results.Add(new Result("Hypervolume", typeof(DoubleValue)));
      }
      if (!results.ContainsKey("BestKnownHypervolume")) {
        results.Add(new Result("BestKnownHypervolume", typeof(DoubleValue)));
      }
      if (!results.ContainsKey("Absolute Distance to BestKnownHypervolume")) {
        results.Add(new Result("Absolute Distance to BestKnownHypervolume", typeof(DoubleValue)));
      }
      Hypervolume comp = new Hypervolume(TestFunction.ReferencePoint, Maximization);
      RealVector[] front = NonDominatedSelect.selectNonDominatedRows(qualities, Maximization, true);

      double hv = comp.GetHypervolume(front);
      double best = TestFunction.BestKnownHypervolume;
      results["Hypervolume"].Value = new DoubleValue(hv);
      results["BestKnownHypervolume"].Value = new DoubleValue(best);
      results["Absolute Distance to BestKnownHypervolume"].Value = new DoubleValue(best - hv);
    }

    [StorableConstructor]
    private MultiObjectiveTestFunctionProblem(bool deserializing) : base(deserializing) { }
    private MultiObjectiveTestFunctionProblem(MultiObjectiveTestFunctionProblem original, Cloner cloner)
      : base(original, cloner) {
      RegisterEventHandlers();
    }
    public MultiObjectiveTestFunctionProblem()
      : base() {
      Parameters.Add(new FixedValueParameter<IntValue>("ProblemSize", "The dimensionality of the problem instance (number of variables in the function).", new IntValue(2)));
      Parameters.Add(new FixedValueParameter<IntValue>("SolutionSize", "The dimensionality of the solution vector (number of objectives).", new IntValue(2)));
      Parameters.Add(new ValueParameter<DoubleMatrix>("Bounds", "The bounds of the solution given as either one line for all variables or a line for each variable. The first column specifies lower bound, the second upper bound.", new DoubleMatrix(new double[,] { { -4, 4 } })));
      Parameters.Add(new ValueParameter<IMultiObjectiveTestFunction>("TestFunction", "The function that is to be optimized.", new Fonseca()));

      Encoding.LengthParameter = ProblemSizeParameter;
      Encoding.BoundsParameter = BoundsParameter;

      InitializeOperators();
      RegisterEventHandlers();
    }



    public override IDeepCloneable Clone(Cloner cloner) {
      return new MultiObjectiveTestFunctionProblem(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEventHandlers();
    }

    private void RegisterEventHandlers() {
      TestFunctionParameter.ValueChanged += TestFunctionParameterOnValueChanged;
      ProblemSizeParameter.Value.ValueChanged += ProblemSizeOnValueChanged;
      SolutionSizeParameter.Value.ValueChanged += SolutionSizeOnValueChanged;
      BoundsParameter.ValueChanged += BoundsParameterOnValueChanged;
    }

    public double[] Evaluate(RealVector individual, IRandom random) {
      return TestFunction.Evaluate(individual);
    }

    public override double[] Evaluate(Individual individual, IRandom random) {
      return Evaluate(individual.RealVector(), random);
    }


    #region Events
    protected override void OnEncodingChanged() {
      base.OnEncodingChanged();
      Parameterize();
    }
    protected override void OnEvaluatorChanged() {
      base.OnEvaluatorChanged();
      Parameterize();
    }

    private void TestFunctionParameterOnValueChanged(object sender, EventArgs eventArgs) {
      var problemSizeChange = ProblemSize < TestFunction.MinimumProblemSize
                              || ProblemSize > TestFunction.MaximumProblemSize;
      if (problemSizeChange) {
        ProblemSize = Math.Max(TestFunction.MinimumProblemSize, Math.Min(ProblemSize, TestFunction.MaximumProblemSize));
      }

      var solutionSizeChange = SolutionSize < TestFunction.MinimumSolutionSize
                              || SolutionSize > TestFunction.MaximumSolutionSize;
      if (solutionSizeChange) {
        ProblemSize = Math.Max(TestFunction.MinimumSolutionSize, Math.Min(SolutionSize, TestFunction.MaximumSolutionSize));
      }
      Bounds = (DoubleMatrix)TestFunction.Bounds.Clone();
      OnReset();
    }

    private void ProblemSizeOnValueChanged(object sender, EventArgs eventArgs) {
      if (ProblemSize < TestFunction.MinimumProblemSize
        || ProblemSize > TestFunction.MaximumProblemSize)
        ProblemSize = Math.Min(TestFunction.MaximumProblemSize, Math.Max(TestFunction.MinimumProblemSize, ProblemSize));
      if (SolutionSize < TestFunction.MinimumSolutionSize
        || SolutionSize > TestFunction.MaximumSolutionSize)
        SolutionSize = Math.Min(TestFunction.MaximumSolutionSize, Math.Max(TestFunction.MinimumSolutionSize, SolutionSize));
      TestFunction.ActualSolutionSize = SolutionSize;
    }

    private void SolutionSizeOnValueChanged(object sender, EventArgs eventArgs) {
      if (SolutionSize < TestFunction.MinimumSolutionSize
        || SolutionSize > TestFunction.MaximumSolutionSize)
        SolutionSize = Math.Min(TestFunction.MaximumSolutionSize, Math.Max(TestFunction.MinimumSolutionSize, SolutionSize));
      TestFunction.ActualSolutionSize = SolutionSize;
      if (ProblemSize < TestFunction.MinimumProblemSize
        || ProblemSize > TestFunction.MaximumProblemSize)
        ProblemSize = Math.Min(TestFunction.MaximumProblemSize, Math.Max(TestFunction.MinimumProblemSize, ProblemSize));
    }

    private void BoundsParameterOnValueChanged(object sender, EventArgs eventArgs) {
      Parameterize();
    }
    #endregion

    #region Helpers
    private void InitializeOperators() {
      //empty for now
      Parameterize();
    }

    private void Parameterize() {
      //empty for now
    }

    #endregion
  }
}

